(* Mathematica Package *)

(* Created by the Wolfram Workbench Nov 3, 2014 *)

BeginPackage["WildfiresDoc`"]
(* Exported symbols added here with SymbolName::usage *) 

(* --- Exposed RBF symbols. --------*)
RadialBasicFunction;
RBFInterpolation;

(* Messages *)
RadialBasicFunction::nnrbf = "Radial basic function `1` not found. Returning default basic function.";
RadialBasicFunction::nnfun = "Radial basic function `1` is not of the form Function[{e,r},...]. Returning default basic function.";
RadialBasicFunction::nnform = "Radial basic function `1` is either not a valid basic function name or does not have the correct form. Returning default basic function.";

(* --- Exposed RBF Interpolation symbols ---*)
Options[RBFInterpolation] = {"Solver" -> Automatic, "Regularization"->0.};

RBFInterpolation;

RBFInterpolation::scale = "The specified scaling parameter was not numeric. The scaling parameter will be computed automatically.";
RBFInterpolation::vscale = "The list of specified scaling parameters were not numeric or not the same length as data. The scaling parameter will be computed automatically";
RBFInterpolation::reg = "The regularization is not a number or a numeric vector of same size as input data points. Regularization will not be used.";
RBFInterpolation::solver = "`1` is not one of LinearSolve, LeastSquares, RileySolve. Solver will be chosen automatically.";
RBFInterpolation::nndata = "The input data was not in the form {{x1,y1,..},val1}, {x2,y2,...},val2},...} or was not numeric.";

(* --- Exposed Divergence-free RBF 2D symbols ---*)
Options[DivergenceFreeInterpolation2D] = {"Solver" -> Automatic, "Regularization"-> 0.};

DivergenceFreeInterpolation2D;

DFInterpolation::coords2d = "The input coordinates are not two-dimensional. The Function only performs 2D interpolation.";
DFInterpolation::nndata2d = "The input data are not two-dimensional. At each of the coordinates, the data should be of the form {u,v}.";

(* --- Exposed Divergence-free RBF 3D symbols ---*)
Options[DivergenceFreeInterpolation3D] = {"Solver" -> Automatic, "Regularization"-> 0.};
DivergenceFreeInterpolation3D;

DFInterpolation3D::coords3d = "The input coordinates are not three-dimensional. The Function only performs 3D interpolation.";
DFInterpolation3D::nndata3d = "The input data are not three-dimensional. At each of the coordinates, the data should be of the form {u,v, w}.";

(* --- Exposed SDGE function -----*)
LoadSDGEWeatherStationInfo;
LoadSDGEWeatherData;
LoadSDGESensorLocations;
LoadSDGEVelocityData;

(* --- Exposed conversion function ----*)
ConvertPositionFromCartesianToSigma;
ConvertPositionFromSigmaToCartesian;
ConvertVelocityFromCartesianToSigma ; 
ConvertVelocityFromSigmaToCartesian ; 

(* ---- Exposed level-set transport solver functions ----*)
Options[TrackLevelSet] = {"Diffusion" -> Automatic, "nGridPoints" -> Automatic, "LengthScale"-> Automatic};
TrackLevelSet;

(* --- Exposed Utility function symbols ---*)
Extract2DIsoSurfacePoints;
Wildfires`Utitlity`VelocityDataFromTerrainAndBCs ; 
Wildfires`Utility`VelocityDataInXYPlane;
Wildfires`Utility`DiscretizeElevationMap;
Wildfires`Utility`TestCase;
Wildfires`Utility`PlotZeroLevelSetsInSigmaPlane;
Wildfires`Utility`PlotZeroLevelSetsInCartesianDomain;

$SDGEDATAPATH =  "/Users/paritoshm/Documents/SDGEData";

Begin["`Private`"]
(* Implementation of the package *)


(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 								INTERNAL UTILITY FUNCTIONS												*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

$RadiusOfEarthInFeet = 7958750000/381;
$da = Pi/180;
$eps = 10^-16.; 

MatrixNumericQ[points_]:= MatrixQ[points, NumericQ];
VectorNumericQ[points_]:= VectorQ[points, NumericQ];
VectorIntegerQ[points_]:= VectorQ[points, IntegerQ];
MatrixColumnIs3[points_]:= TrueQ[Last[Dimensions[points]] == 3];
MatrixColumnIs2[points_]:= TrueQ[Last[Dimensions[points]] == 2];

MakeListableCompiledFunction[f_, v_Symbol] := 
	With[{p2 = {v}}, Compile[{{p, _Real}}, Block[p2, v = p; f],
	RuntimeAttributes -> {Listable}, RuntimeOptions -> {"EvaluateSymbolically" -> False, "RuntimeErrorHandler"->Function[$Failed]}]];

MakeListableCompiledFunction[f_, vars_List] := Block[{v = Flatten[List[vars]]},
	If[Length[v] == 1,
		MakeListableCompiledFunction[f, v[[1]] ]
		,
		Compile[{{p, _Real, 1}}, Block[vars, vars = p; f], RuntimeAttributes -> {Listable}, 
			RuntimeOptions -> {"EvaluateSymbolically" -> False, "RuntimeErrorHandler"->Function[$Failed]}]
	]
];


RileySolve[mat_, rhs_] := Module[{L, LU, LL, z, y, x, mu = 10^-11},
   L = CholeskyDecomposition[mat + mu IdentityMatrix[Length[mat], SparseArray]]; 
   LL = LinearSolve[Transpose[L]]; LU = LinearSolve[L];
   Remove[L];
   z = LL[rhs];
   y = x = LU[z];
   Do[
   	z = mu LL[y]; y = LU[z];
    x += y;
    , {5}];
   x
   ];
   
PiecewiseToIf[f_] := If[FreeQ[f, Piecewise], f, f /. Piecewise[{{a_, b_}}] :> If[b, a, 0]];
PiecewisePresentQ[f_] := Not[FreeQ[f, Piecewise]];

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR GENERATING AND CHECKING RADIAL BASIC FUNCTIONS					*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

$DefaultRBF = "Gaussian";

RBF[name_String] := Switch[name,
   "Gaussian", Function[{e, r}, Exp[-(e r)^2]], 
   "MultiQuadric" | "MQ", Function[{e, r}, Sqrt[1 + (e r)^2]], 
   "InverseMultiQuadric" | "IMQ", Function[{e, r}, 1/Sqrt[1 + (e r)^2]],  
   "Radial", Function[{e, r}, r],
   "GeneralizedInverseMultiQuadric" | "GIMQ", Function[{e, r}, 1/(1 + (e r)^2)^2], 
   "InverseQuadric" | "IQ", Function[{e, r}, 1/(1 + (e r)^2)],
   "Matern", Function[{e, r}, Exp[-e r]],
   "LinearMatern", Function[{e, r}, Exp[-e r] (1 + e r)],
   "QuadraticMatern", Function[{e, r}, Exp[-e r] (3 + 3 e r + (e r)^2)],
   "CubicMatern", Function[{e, r}, Exp[-e r] (15 + 15 (e r) + 6 (e r)^2 + (e r)^3)],
   "GeneralizedMultiQuadric2" | "GMQ2", Function[{e, r}, (1 + (e r)^2)^(3/2)],
   "Cubic", Function[{e, r}, r^3],
   "Polyharmonic", Function[{e, r}, (r) Log[r+$eps]],
   "ThinPlateSpline" | "TPS", Function[{e, r}, (r)^2 Log[r+$eps]],
   "TPS2" | "ThinPlateSplineOrder2", Function[{e, r}, r^4 Log[r+$eps]],
   "TPS4" | "ThinPlateSplineOrder4", Function[{e,r}, r^8 Log[r + $eps]],
   "GeneralizedMQ3", Function[{e, r}, (1 + (e r)^2)^(5/2)],
   "Qunitic", Function[{e, r}, r^5],
   "Septic", Function[{e, r}, r^7],
   "LinearLaguerreGaussian" | "LLG", Function[{e, r}, Exp[-(e r)^2] (2 - (e r)^2)],
   "QuadraticLaguerreGaussian" | "QLG", Function[{e, r}, Exp[-(e r)^2] (3 - 3 (e r)^2 + ((e r)^4)/2)],
   "LinearGenralizedIMQ" | "LGIMQ", Function[{e, r}, (2 - (e r)^2)/(1 + (e r)^2)^4],
   "QuadraticGeneralizedIMQ" | "QGIMQ", Function[{e, r}, (3 - 6 (e r)^2 + (e r)^4)/(1 + (e r)^2)^6],
   "Wendland30", Function[{e, r}, Piecewise[{{(1 - e r)^2, e r <= 1}}]],
   "Wendland31", Function[{e, r}, (4 e r + 1) Piecewise[{{(1 - e r)^4, e r <= 1}}]],
   "Wendland32", Function[{e, r}, (35 (e r)^2 + 18 (e r) + 3) Piecewise[{{(1 - e r)^6, e r <= 1}}]],
   "Wendland33", Function[{e, r}, (32 (e r)^3 + 25 (e r)^2 + 8 (e r) + 1) Piecewise[{{(1 - e r)^8, e r <= 1}}]],
   "Wu33", Function[{e, r}, (5 (e r)^3 + 20 (e r)^2 + 29 (e r) + 16) Piecewise[{{(1 - e r)^4, e r <= 1}}]],
   "Wu23", Function[{e, r}, (5 (e r)^4 + 25 (e r)^3 + 48 (e r)^2 + 40 (e r) + 8) Piecewise[{{(1 - e r)^5, e r <= 1}}]],
   "Wu13", Function[{e, r}, (5 (e r)^5 + 30 (e r)^4 + 72 (e r)^3 + 82 (e r)^2 + 36 (e r) + 6) Piecewise[{{(1 - e r)^6, e r <= 1}}]],
   "Wu03", Function[{e, r}, (5 (e r)^6 + 35 (e r)^5 + 101 (e r)^4 + 147 (e r)^3 + 101 (e r)^2 + 35 (e r) + 5) Piecewise[{{(1 - e r)^7, e r <= 1}}]],
   "IntegratedMultiQuadric1" | "IntegratedMQ1", Function[{e,r}, (r*Sqrt[1 + (e*r)^2])/2 + ArcSinh[e*r]/(2*e)], 
   "IntegratedMultiQuadric2" | "IntegratedMQ2", Function[{e,r}, (-(Sqrt[1 + (e*r)^2]/e) + (1 + (e*r)^2)^(3/2)/(3*e) + r*ArcSinh[e*r])/(2*e)], 
   "IntegratedInverseMultiQuadric1" | "IntegratedMQ1", Function[{e,r}, ArcSinh[e*r]/e], 
   "IntegratedInverseMultiQuadric2" | "IntegratedMQ2", Function[{e,r}, (-(Sqrt[1 + (e*r)^2]/e) + r*ArcSinh[e*r])/e],
   "Automatic", RBF[$DefaultRBF],
   _, $Failed
   ];
RBF[Automatic] := RBF[$DefaultRBF];
RBF[___]:= $Failed;

RBFType["Names"] = {"Gaussian", {"MultiQuadric", "MQ"}, {"InverseMultiQuadric", "IMQ"}, {"Polyharmoninc", "TPS"}, 
   "Radial",{"GeneralizedInverseMultiQuadric" , "GIMQ"}, {"InverseQuadric" , "IQ"}, "Matern", "LinearMatern", "QuadraticMatern", 
   "CubicMatern", {"GeneralizedMultiQuadric2" , "GMQ2"}, {"GeneralizedMultiQuadric3", "GMQ3"}, "Cubic", 
   "ThinPlateSpline", "Qunitic", { "ThinPlateSplineOrder2", "TPS2"}, {"TPS4" , "ThinPlateSplineOrder4"},"Septic", {"LinearLaguerreGaussian" , 
    "LLG"}, {"QuadraticLaguerreGaussian" , "QLG"}, {"LinearGenralizedIMQ" , "LGIMQ"},
   {"QuadraticGeneralizedIMQ" , "QGIMQ"}, "Wendland30", "Wendland31", 
   "Wendland32", "Wendland33", "Wu33", "Wu23", "Wu13", "Wu03", {"IntegratedMultiQuadric1" , "IntegratedMQ1"}, {"IntegratedMultiQuadric2" , "IntegratedMQ2"},
   {"IntegratedInverseMultiQuadric1" , "IntegratedInverseMQ1"}, {"IntegratedInverseMultiQuadric2" , "IntegratedInverseMQ2"}};

GetRadialBasicFunction[name_String]:= With[{rbf = RBF[name]}, If[rbf === $Failed, "NotAvailable", Evaluate[rbf[#1,#2]]& ]] ; 

iRadialBasicFunctionString[rbf_String, defaultrbf_String]:= Module[{basicfun},
	If[rbf === "Names", Return[RBFType["Names"]] ];
	basicfun = GetRadialBasicFunction[rbf];
	If[basicfun === "NotAvailable",
		Message[RadialBasicFunction::nnrbf, rbf];
		basicfun = GetRadialBasicFunction[defaultrbf] 
	];
	basicfun
];
iRadialBasicFunctionFunction[rbf_Function, defaultrbf_String]:= Module[{fi, basicfun},
	fi = Flatten[List[First[rbf]]];
	If[Length[fi] !=2, 
		Message[RadialBasicFunction::nnfun, rbf];
		Return[GetRadialBasicFunction[defaultrbf]];
	];
	basicfun = Function@@List[fi, Last[rbf]] ; 
	basicfun = Evaluate[basicfun[#1,#2]]&;
	If[Not[NumericQ[basicfun[1.,1.]]], 
		Message[RadialBasicFunction::nnform, rbf];
		basicfun = GetRadialBasicFunction[defaultrbf];
	];
	basicfun
];
iRadialBasicFunction[rbf_, defaultrbf_String]:= Module[{sname},
	sname = If[StringQ[rbf], iRadialBasicFunctionString, iRadialBasicFunctionFunction];
	sname[rbf, defaultrbf]
]; 

RadialBasicFunction[Automatic] := RBF[Automatic];
RadialBasicFunction["Automatic"] := RBF[Automatic];
RadialBasicFunction[rbf_String] := iRadialBasicFunctionString[rbf, $DefaultRBF];
RadialBasicFunction[rbf_Function]:= iRadialBasicFunctionFunction[rbf, $DefaultRBF];
RadialBasicFunction[p___]:= (Message[RadialBasicFunction::nnform, p]; GetRadialBasicFunction[$DefaultRBF]);
	
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR PERFORMING RBF INTERPOLATION							*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

RBFDistanceMatrix[evalpts_List, centers_List]:= Block[{nc = Length[centers], ec = Length[evalpts]},
	Sqrt[Transpose[ConstantArray[Map[Total, evalpts^2], nc]] + ConstantArray[Map[Total, centers^2], ec] - 2*evalpts.Transpose[centers]]
];

RBFInterpolationMatrix[centers_List, rbf_Function, scales_List, regularization_?NumericQ]:= Module[{sm, dm},
	If[Not[VectorNumericQ[scales]], Return[$Failed]];
	dm = RBFDistanceMatrix[centers,centers];
	sm = If[Length[scales]==Length[centers], SparseArray[ConstantArray[scales, Length[centers]] ], First[scales]];
	rbf[sm, dm] + regularization*IdentityMatrix[Length[centers], SparseArray]
];

RBFInterpolationMatrixSparse[centers_List, rbf_Function, scales_List, regularization_?NumericQ]:= Module[{sm, fun, im, ci, temp},
	If[Not[VectorNumericQ[scales]], Return[$Failed]];
	sm = If[Length[scales]==1, First[scales], scales] ;
	fun = PiecewiseToIf[rbf];
	im = Table[
		temp = Flatten@RBFDistanceMatrix[List[ci], centers];
		SparseArray[List[If[NumericQ[sm], Map[fun[sm, #] &, temp], MapThread[fun, {sm, temp}]]] ]
		,{ci, centers}
	];
	im = Join@@im;
	im + regularization*IdentityMatrix[Length[centers],SparseArray]	
];

RBFBases[centers_List, scales_List, rbf_Function]:= Module[{cv, vars, dim, dm, sm, fun, bases},
	dim = Length[First[centers]];
	vars = Table[Unique[cv], {dim}];
	dm = Flatten[RBFDistanceMatrix[{vars}, centers]];
	sm = If[Length[scales]==1, First[scales], scales];
	fun = PiecewiseToIf[rbf];
	bases = If[PiecewisePresentQ[rbf], 
		If[NumericQ[sm], Map[fun[sm,#]&, dm], MapThread[fun, {sm,dm}]]
		,
		rbf[sm,dm]
	];
	{vars, dim, bases}
];

RBFInterpolationValue[nf_Experimental`NumericalFunction, inputdim_Integer, query_]:= Module[{},
	If[inputdim==1,
		If[VectorNumericQ[query], Return[Map[nf[List[#]]&, query]] ];
		If[NumericQ[query], Return[nf[{query}]] ];
		If[MatrixNumericQ[query], Return[If[Length[First[query]] == 1, Map[nf, query], $Failed]] ];
		Return[$Failed];	
	];
	If[VectorNumericQ[query], Return[If[Length[query] == inputdim, nf[query], $Failed]] ];
	If[MatrixNumericQ[query], Return[Map[nf,query]] ];
	$Failed
];

RBFInterpolatingFunctionDerivative[dfun_, derivinfo_]:= Module[{ivars, derivs, dfun0, nf0},
	If[Not[VectorIntegerQ[derivinfo]] || Length[derivinfo] != dfun["InputDimensions"], Return[$Failed]];
	ivars = dfun["NumericalFunction"]["ArgumentNames"] ; 
	derivs = Transpose[{ivars, derivinfo}];
	nf0 = Experimental`CreateNumericalFunction@@List[ivars, D[dfun["NumericalFunction"]["FunctionExpression"], Sequence@@derivs], {}]; 
	If[Not[NDSolve`ValidNumericalFunctionQ[nf0]], Return[$Failed]];
	
	(* ---- Generate the black-box function with certain properties ----*)
   	dfun0["InputDimensions"] = dfun["InputDimensions"];
   	dfun0["Coefficients"] = dfun["Coefficients"];
   	dfun0["BasicFunction"] = "NotAvailable";
   	dfun0["Scales"] = dfun["Scales"];
   	dfun0["NumericalFunction"] = nf0;
   	dfun0["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun0[query__] := RBFInterpolationValue[dfun0["NumericalFunction"], dfun0["InputDimensions"], query];
   	Derivative /: Derivative[deriv___][dfun0] := RBFInterpolatingFunctionDerivative[dfun0, List[deriv]];
   	MakeBoxes[dfun0, _] ^:= InterpretationBox["RBFInterpolatingFunction[<>]", dfun0];
	
	Derivative /: Derivative[Sequence@@derivinfo][dfun] = dfun0;
	dfun0
];

FitRBF[centers_List, scales_List, fval_, rbf_Function, regularization_, soltype_] := Module[{sname, mat, coeffs, bases, outputfun, dfun, vars,dim},
	
	sname = If[PiecewisePresentQ[rbf], RBFInterpolationMatrixSparse, RBFInterpolationMatrix];
	
	mat = sname[centers, rbf, scales, First[Flatten[List[regularization]]]];
	sname = Switch[soltype, "Riley", RileySolve, "LinearSolve" | LinearSolve, LinearSolve, "LeastSquares" | LeastSquares, 
   	LeastSquares, _, LeastSquares];
   	
   	coeffs = Quiet@sname[mat, fval];
   	If[Not[VectorNumericQ[coeffs]], Return[$Failed]];
   	
   	{vars, dim, bases} = RBFBases[centers, scales, rbf];
   	(*outputfun = MakeListableCompiledFunction[coeffs.bases, vars] ;*)
   	outputfun = Experimental`CreateNumericalFunction[vars, coeffs.bases,{}]; 
   
   	(* ---- Generate the black-box function with certain properties ----*)
   	dfun["InputDimensions"] = dim;
   	dfun["Coefficients"] = coeffs;
   	dfun["BasicFunction"] = rbf;
   	dfun["Scales"] = scales;
   	dfun["NumericalFunction"] = outputfun;
   	dfun["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun[query__] := RBFInterpolationValue[dfun["NumericalFunction"], dfun["InputDimensions"], query];
   	Derivative /: Derivative[deriv___][dfun] := RBFInterpolatingFunctionDerivative[dfun, List[deriv]];
   	MakeBoxes[dfun, _] ^:= InterpretationBox["RBFInterpolatingFunction[<>]", dfun];
   	dfun
   ];

GetVectorScalingParameters[centers_, rbf_, oscales_List, ratio_?NumericQ]:= Module[{factor, sol,e, r, nf},
	factor = factor = 1 + Sign[rbf[1, 1] - rbf[1, 10^-9]]*ratio ; 
	sol = e /. Quiet[Solve[rbf[e,r]==factor, e]] ; 
	e = Function@@List[{r}, If[sol === e, 1, Abs[First[sol]] ]];
	nf = Nearest[centers];
	sol = MapThread[If[NumericQ[#1], #1, e[EuclideanDistance@@nf[#2,2]]]&, {oscales, centers}];
	If[Not[VectorNumericQ[sol]], $Failed, sol]
];

GetSingleScalingParameter[centers_List, fval_List, rbf_Function, ratio_?NumericQ] := Module[{nf, minmaxpts, maxDia, maxR, e, e1, fun},
  nf = Nearest[centers];
  minmaxpts = Transpose[Map[{Min[#], Max[#]} &, Transpose[centers]]] ;
  maxDia = Norm[minmaxpts];
  maxR = Map[Sqrt[Total[(Last[nf[#, {Infinity, 0.2*maxDia}]] - #)^2]] &, centers] ; 
  maxR = maxR /. 0. -> Mean[maxR];
  
  fun = (rbf[e, 10^-12] - rbf[e, Mean[maxR]])^2 - ratio^2;
  fun = PiecewiseToIf[fun];
  e1 =  List[e] /. Quiet@FindRoot[Evaluate[fun], {e, 10^-3}];
  Abs[e1]
  ];

GetRBFScalingParameter[centers_List, fval_List, rbf_Function, oscales_List, ratio_]:= Module[{scales = oscales},
	If[Length[oscales]==1, 
		If[MatchQ[First[scales], Automatic | "Automatic"], Return[GetSingleScalingParameter[centers, fval, rbf, ratio]] ];
		If[Not[VectorNumericQ[scales]], 
			Message[RBFInterpolation::scale, scales];
			Return[GetSingleScalingParameter[centers, fval, rbf, ratio]];
		];
		Return[scales];
	];
	If[Not[VectorNumericQ[scales]] || Length[scales] != Length[centers], 
		Message[RBFInterpolation::vscale];
		Return[GetSingleScalingParameter[centers, fval, rbf, ratio]]
		,
		Return[Flatten[scales]]
	]; 
	GetSingleScalingParameter[centers, fval, rbf, ratio]
];


iRBFInterpolation[data_List, rbf_Function, oscales_, opts_]:= Module[{centers, fval, regularization, solvertype, datalen, reglen, scales, 
	inputlen, outputlen, temp, temp2},
	
	temp = Map[Flatten, data];
	If[Not[MatrixNumericQ[temp]], Return[$Failed]];
	temp2 = data[[1]];
	If[Length[temp2] != 2,
		Message[RBFInterpolation::nndata]; 
		Return[$Failed]
	];
	inputlen = Length[Flatten[List[First[temp2]]]];
	outputlen = Length[Flatten[List[Last[temp2]]]];
	If[outputlen > 1, Return[$Failed]];
	centers = temp[[All,1;;inputlen]];
	fval = Flatten[temp[[All,inputlen+1;;-1]]];
	{centers,fval} = Developer`ToPackedArray /@ {centers, fval};
	
	If[Not[MatrixNumericQ[centers]] || Not[VectorNumericQ[fval]], 
		Message[RBFInterpolation::nndata];
		Return[$Failed]
	];
	datalen = Length[centers];
	{regularization, solvertype} = {"Regularization", "Solver"} /. Join[opts, Options[RBFInterpolation]];
	scales = GetRBFScalingParameter[centers, {}, rbf, Flatten[List[oscales]], 0.2];
	If[scales === $Failed, Return[$Failed]];
	
	(* --- Check regularization option --*)
	regularization = Flatten[List[regularization]];
	reglen = Length[regularization];
	If[reglen != 1 && reglen != datalen, 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Length[regularization]==1  && Not[NumericQ[regularization[[1]]]], 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Not[VectorNumericQ[regularization]], Message[rbf::noreg]; regularization = {0}];
	
	(* --- Check solver option --*)
	solvertype = First[Flatten[List[solvertype]]];
	If[MatchQ[solvertype, Automatic|"Automatic"], solvertype = "LinearSolve"];
	If[Not[MatchQ[solvertype, "RileySolve" | "LinearSolve" | "LeastSquares"]], 
		Message[RBFInterpolation::solver, solvertype];
		solvertype = If[Length[scales]==1, "RileySolve", "LinearSolve"]
	];
	If[Length[scales]>1 && solvertype === "RileySolve", solvertype = "LinearSolve"];
	
	
	FitRBF[centers, scales, fval, rbf, regularization, solvertype]
];

RBFInterpolation[datapts_, opts:OptionsPattern[]] := iRBFInterpolation[N@datapts, RadialBasicFunction[$DefaultRBF], Automatic, List[opts]];
RBFInterpolation[datapts_, basicfun_, opts:OptionsPattern[]] := iRBFInterpolation[N@datapts, RadialBasicFunction[basicfun], Automatic, List[opts]]; 
RBFInterpolation[datapts_, basicfun_, rbfscales_, opts:OptionsPattern[]] :=iRBFInterpolation[N@datapts, RadialBasicFunction[basicfun], rbfscales, List[opts]];
RBFInterpolation[___]:= $Failed;

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR PERFORMING 2D DIV.FREE RBF INTERPOLATION							*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
 
DivergenceFreeVectorRBF2D[basicfun_Function, xy_List, centers_List]:= Block[{r, p, gI, e},
	If[Not[Length[Flatten[xy]]==2 || Length[Flatten[centers]]==2 || VectorQ[xy, MatchQ[#, _Symbol] &]
		|| VectorQ[centers, MatchQ[#, _Symbol] &]], Return[$Failed]];
	r = Sqrt[Total[(xy - centers)^2]];
	p = basicfun[e,r];
	gI = With[{gg = -Laplacian[p, xy]}, {{gg, 0}, {0, gg}}];
	gI = gI + Outer[D[p, #1, #2] &, xy, xy] ;
	Map[Function@@List[Flatten[{e, xy, centers}], #]&, gI]
];

DivergenceFreeVectorBases2D[basicfun_, centers_List, scales_]:= Block[{ubases, vbases, x,y, xc, yc,e},
	If[Not[MatrixNumericQ[centers] || VectorNumericQ[scales]], Return[$Failed]];
	{ubases, vbases} = DivergenceFreeVectorRBF2D[basicfun, {x,y}, {xc,yc}];
	
	e = Flatten[List[scales]];
	If[Not[VectorNumericQ[e]], Return[$Failed]];
	If[Length[e] ==1, e = ConstantArray[First[e], Length[centers]] ];
	If[Length[e] != Length[centers], Return[$Failed]];
	
	ubases = Flatten[MapThread[ubases@@Flatten[List[#1,{x, y}, #2]]&, {e, centers}]];
	vbases = Flatten[MapThread[vbases@@Flatten[List[#1,{x, y}, #2]]&, {e, centers}]];
	
	ubases = Function@@List[{x,y}, ubases];	
	vbases = Function@@List[{x,y}, vbases];
	{ubases, vbases}
];

DivFreeRBF2DInterpolantDerivative[dfun_, {mm_, nn_}] := Module[{dfun0, nf0, ivars, deriv},
	ivars = First[dfun["NumericalFunction"]]["ArgumentNames"] ; 
	deriv = Transpose[{ivars, {mm,nn}}];
	nf0 = Map[Experimental`CreateNumericalFunction@@List[ivars, D[#["FunctionExpression"], Sequence@@deriv], {}]&, dfun["NumericalFunction"]]; 
	If[Not[And@@Map[NDSolve`ValidNumericalFunctionQ[#]&, nf0]], Return[$Failed]];
	
	(* ---- Generate the black-box function with certain properties ----*)
   	dfun0["InputDimensions"] = 2;
   	dfun0["Coefficients"] = dfun["Coefficients"];
   	dfun0["BasicFunction"] = "NotAvailable";
   	dfun0["Scales"] = dfun["Scales"];
   	dfun0["NumericalFunction"] = nf0;
   	dfun0["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun0[{x$_?NumericQ, y$_?NumericQ}] := Through[dfun0["NumericalFunction"][{x$,y$}]];
   	dfun0[xy$$ : {{_?NumericQ, _?NumericQ} ...}] := Map[dfun0, xy$$]; 
   	Derivative /: Derivative[m_?IntegerQ, n_?IntegerQ][dfun0] := DivFreeRBF2DInterpolantDerivative[dfun0, {m,n}]; 
   	
   	Derivative /: Derivative[mm, nn][dfun] = dfun0;
   	MakeBoxes[dfun0, _] ^:= InterpretationBox["DivergenceFreeInterpolatingFunction[<>]", dfun0];
   	dfun0
];

FitDivFreeRBF2D[centers_List, scales_List, fval_List, rbf_Function, regularization_, solvertype_] := Module[{reg, temp,
	ub,vb, uc, vc, umat, vmat, finalmat, rhs, coeffs, nf, ui,x,y, sname, dfun},
	reg = First[Flatten[List[regularization]]];
	temp = DivergenceFreeVectorBases2D[rbf, centers, scales];
	If[temp === $Failed, Return[$Failed], {ub, vb} = temp];
	
	{uc, vc} = Map[MakeListableCompiledFunction[#[x,y],{x,y}]&, {ub,vb}];
	{umat, vmat} = Map[#[centers+2*$eps]&, {uc, vc}];
	Map[If[Not[MatrixNumericQ[#]], Return[$Failed]]&, {umat, vmat}];
	
	finalmat = Join[umat, vmat] + reg*IdentityMatrix[2*Length[umat],SparseArray];
	rhs = Flatten[Transpose[fval]];
	If[Length[finalmat] != Length[rhs], Return[$Failed]];
	
	sname = Switch[solvertype, "LinearSolve" | LinearSolve, LinearSolve, "LeastSquares" | LeastSquares, LeastSquares, _, LeastSquares];
	coeffs = Quiet@sname[finalmat, rhs];
	
	If[Not[VectorNumericQ[coeffs]], Return[$Failed]];
	nf = Table[
		Experimental`CreateNumericalFunction@@List[{x,y}, ui[x,y].coeffs, {}]
		,{ui, {ub, vb}}
	];
	If[Not[And@@Map[NDSolve`ValidNumericalFunctionQ, nf]], Return[$Failed]];
	Remove[ub,vb,uc,vc,umat,vmat, finalmat,rhs, temp];
	
	(* ---- Generate the black-box function with certain properties ----*)
   	dfun["InputDimensions"] = 2;
   	dfun["Coefficients"] = coeffs;
   	dfun["BasicFunction"] = rbf;
   	dfun["Scales"] = scales;
   	dfun["NumericalFunction"] = nf;
   	dfun["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun[{x$_?NumericQ, y$_?NumericQ}] := Through[(dfun["NumericalFunction"])[{x$,y$}]];
   	dfun[xy$$ : {{_?NumericQ, _?NumericQ} ...}] := Map[dfun, xy$$]; 
   	
   	Derivative /: Derivative[m_?IntegerQ, n_?IntegerQ][dfun] := DivFreeRBF2DInterpolantDerivative[dfun, {m,n}]; 
   	MakeBoxes[dfun, _] ^:= InterpretationBox["DivergenceFreeInterpolatingFunction[<>]", dfun];
   	dfun
]

iDFRBFInterpolation2D[data_, rbf_Function, oscales_, opts_]:= Module[{temp, temp2, inputlen, outputlen, centers, fval, datalen, regularization,
	solvertype, scales, reglen},
	temp = Map[Flatten, data];
	If[Not[MatrixNumericQ[temp]], Return[$Failed]];
	temp2 = data[[1]];
	If[Length[temp2] != 2,
		Message[RBFInterpolation::nndata]; 
		Return[$Failed]
	];
	
	inputlen = Length[Flatten[List[First[temp2]]]];
	If[inputlen != 2, Message[DFInterpolation::coords2D]; Return[$Failed]];
	
	outputlen = Length[Flatten[List[Last[temp2]]]];
	If[outputlen != 2, Message[DFInterpolation::nndata2D]; Return[$Failed]];
	
	centers = temp[[All,1;;inputlen]];
	fval = temp[[All,inputlen+1;;-1]];
	{centers,fval} = Developer`ToPackedArray /@ {centers, fval};
	
	If[Not[MatrixNumericQ[centers]] Not[MatrixColumnIs2[centers]] || Not[MatrixNumericQ[fval] || Not[MatrixColumnIs2[fval]] ], 
		Message[RBFInterpolation::nndata];
		Return[$Failed]
	];
	datalen = Length[centers];
	{regularization, solvertype} = {"Regularization", "Solver"} /. Join[opts, Options[RBFInterpolation]];
	scales = GetRBFScalingParameter[centers, {}, rbf, Flatten[List[oscales]], 0.003];
	If[scales === $Failed, Return[$Failed]];
	
	(* --- Check regularization option --*)
	regularization = Flatten[List[regularization]];
	reglen = Length[regularization];
	If[reglen != 1 && reglen != datalen, 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Length[regularization]==1  && Not[NumericQ[regularization[[1]]]], 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Not[VectorNumericQ[regularization]], Message[rbf::noreg]; regularization = {0}];
	
	(* --- Check solver option --*)
	solvertype = First[Flatten[List[solvertype]]];
	If[MatchQ[solvertype, Automatic|"Automatic"], solvertype = "LinearSolve"];
	If[Not[MatchQ[solvertype, "LinearSolve" | "LeastSquares"]], 
		Message[RBFInterpolation::solver, solvertype];
		solvertype = "LinearSolve";
	];
	
	FitDivFreeRBF2D[centers, scales, fval, rbf, regularization, solvertype]
];


DivergenceFreeInterpolation2D[datapts_, opts:OptionsPattern[]] := 
								iDFRBFInterpolation2D[N@datapts, RadialBasicFunction[$DefaultRBF], Automatic, List[opts]];
DivergenceFreeInterpolation2D[datapts_, basicfun_, opts:OptionsPattern[]] := 
								iDFRBFInterpolation2D[N@datapts, iRadialBasicFunction[basicfun, "TPS2"], Automatic, List[opts]]; 
DivergenceFreeInterpolation2D[datapts_, basicfun_, rbfscales_, opts:OptionsPattern[]] :=
								iDFRBFInterpolation2D[N@datapts, iRadialBasicFunction[basicfun, "TPS2"], N@rbfscales, List[opts]];
DivergenceFreeInterpolation2D[___]:= $Failed;

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR PERFORMING 3D DIV.FREE RBF INTERPOLATION							*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

DivergenceFreeVectorRBF3D[basicfun_Function, xyz_List, centers_List]:= Module[{r, p, gI, e},
	If[Not[Length[Flatten[xyz]]==3 || Length[Flatten[centers]]==3 || VectorQ[xyz, MatchQ[#, _Symbol] &]
		|| VectorQ[centers, MatchQ[#, _Symbol] &]], Return[$Failed]];
	r = Sqrt[Total[(xyz - centers)^2]];
	p = basicfun[e,r];
	gI = With[{gg = -Laplacian[p, xyz]}, {{gg, 0, 0}, {0, gg, 0}, {0, 0, gg}}];
	gI = gI + Outer[D[p, #1, #2] &, xyz, xyz] ; 
	Map[Function@@List[Flatten[{e, xyz, centers}], #]&, gI]
];

DivergenceFreeVectorBases3D[basicfun_, centers_List, scales_]:= Module[{ubases, vbases, wbases, x,y,z, xc, yc, zc,e},
	If[Not[MatrixNumericQ[centers] || VectorNumericQ[scales]], Return[$Failed]];
	{ubases, vbases, wbases} = DivergenceFreeVectorRBF3D[basicfun, {x,y,z}, {xc,yc,zc}];
	
	e = Flatten[List[scales]];
	If[Not[VectorNumericQ[e]], Return[$Failed]];
	If[Length[e] ==1, e = ConstantArray[First[e], Length[centers]] ];
	If[Length[e] != Length[centers], Return[$Failed]];
	
	ubases = Flatten[MapThread[ubases@@Flatten[List[#1,{x, y, z}, #2]]&, {e, centers}]];
	vbases = Flatten[MapThread[vbases@@Flatten[List[#1,{x, y, z}, #2]]&, {e, centers}]];
	wbases = Flatten[MapThread[wbases@@Flatten[List[#1,{x, y, z}, #2]]&, {e, centers}]];
	
	ubases = Function@@List[{x,y,z}, ubases];	
	vbases = Function@@List[{x,y,z}, vbases];
	wbases = Function@@List[{x,y,z}, wbases];
	{ubases, vbases, wbases}
];

DivFreeRBF3DInterpolantDerivative[dfun_, {mm_, nn_, pp_}] := Module[{dfun0, nf0, ivars, deriv},
	ivars = First[dfun["NumericalFunction"]]["ArgumentNames"] ; 
	deriv = Transpose[{ivars, {mm,nn, pp}}];
	nf0 = Map[Experimental`CreateNumericalFunction@@List[ivars, D[#["FunctionExpression"], Sequence@@deriv], {}]&, dfun["NumericalFunction"]]; 
	If[Not[And@@Map[NDSolve`ValidNumericalFunctionQ[#]&, nf0]], Return[$Failed]];
	
	(* ---- Generate the black-box function with certain properties ----*)
   	dfun0["InputDimensions"] = 3;
   	dfun0["Coefficients"] = dfun["Coefficients"];
   	dfun0["BasicFunction"] = "NotAvailable";
   	dfun0["Scales"] = dfun["Scales"];
   	dfun0["NumericalFunction"] = nf0;
   	dfun0["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun0[{x$_?NumericQ, y$_?NumericQ, z$_?NumericQ}] := Through[dfun0["NumericalFunction"][{x$,y$, z$}]];
   	dfun0[xyz$$ : {{_?NumericQ, _?NumericQ,_?NumericQ} ...}] := Map[dfun0, xyz$$]; 
   	Derivative /: Derivative[m_?IntegerQ, n_?IntegerQ, p_?IntegerQ][dfun0] := DivFreeRBF3DInterpolantDerivative[dfun0, {m,n, p}]; 
   	
   	Derivative /: Derivative[mm, nn, pp][dfun] = dfun0;
   	MakeBoxes[dfun0, _] ^:= InterpretationBox["DivergenceFreeInterpolatingFunction[<>]", dfun0];
   	dfun0
];

FitDivFreeRBF3D[centers_List, scales_List, fval_List, rbf_Function, regularization_, solvertype_] := Module[{reg, temp,
	ub,vb, wb, uc, vc, wc, umat, vmat, wmat, finalmat, rhs, coeffs, nf, ui,x,y, z, sname, dfun},
	reg = First[Flatten[List[regularization]]];
	temp = DivergenceFreeVectorBases3D[rbf, centers, scales];
	If[temp === $Failed, Return[$Failed], {ub, vb, wb} = temp];
	
	{uc, vc, wc} = Map[MakeListableCompiledFunction[#[x,y,z],{x,y,z}]&, {ub,vb, wb}];
	{umat, vmat,wmat} = Map[#[centers+2*$eps]&, {uc, vc, wc}];
	Map[If[Not[MatrixNumericQ[#]], Return[$Failed]]&, {umat, vmat, wmat}];
	
	finalmat = Join[umat, vmat, wmat] + reg*IdentityMatrix[3*Length[umat],SparseArray];
	rhs = Flatten[Transpose[fval]];
	If[Length[finalmat] != Length[rhs], Return[$Failed]];
	
	sname = Switch[solvertype, "LinearSolve" | LinearSolve, LinearSolve, "LeastSquares" | LeastSquares, LeastSquares, _, LeastSquares];
	coeffs = Quiet@sname[finalmat, rhs];
	
	If[Not[VectorNumericQ[coeffs]], Return[$Failed]];
	nf = Table[
		Experimental`CreateNumericalFunction@@List[{x,y,z}, ui[x,y,z].coeffs, {}]
		,{ui, {ub, vb,wb}}
	];
	Map[If[Not[NDSolve`ValidNumericalFunctionQ[#]], Return[$Failed]]&, nf];
	Remove[ub,vb,wb, uc,vc, wc, umat,vmat, wmat, finalmat,rhs, temp];
	
	(* ---- Generate the black-box function with certain properties ----*)
   	dfun["InputDimensions"] = 3;
   	dfun["Coefficients"] = coeffs;
   	dfun["BasicFunction"] = rbf;
   	dfun["Scales"] = scales;
   	dfun["NumericalFunction"] = nf;
   	dfun["Properties"] = {"InputDimensions","Coefficients", "BasicFunction", "Scales", "NumericalFunction"};
   	dfun[{x$_?NumericQ, y$_?NumericQ, z$_?NumericQ}] := Through[(dfun["NumericalFunction"])[{x$,y$, z$}]];
   	dfun[xyz$$ : {{_?NumericQ, _?NumericQ, _?NumericQ} ...}] := Map[dfun, xyz$$]; 
   	
   	Derivative /: Derivative[m_?IntegerQ, n_?IntegerQ, p_?IntegerQ][dfun] := DivFreeRBF2DInterpolantDerivative[dfun, {m,n,p}]; 
   	MakeBoxes[dfun, _] ^:= InterpretationBox["DivergenceFreeInterpolatingFunction[<>]", dfun];
   	dfun
]

iDFRBFInterpolation3D[data_, rbf_Function, oscales_, opts_]:= Module[{temp, temp2, inputlen, outputlen, centers, fval, datalen, regularization,
	solvertype, scales, reglen},
	temp = Map[Flatten, data];
	If[Not[MatrixNumericQ[temp]], Return[$Failed]];
	temp2 = data[[1]];
	If[Length[temp2] != 2,
		Message[DFInterpolation3D::nndata3D]; 
		Return[$Failed]
	];
	
	inputlen = Length[Flatten[List[First[temp2]]]];
	If[inputlen != 3, Message[DFInterpolation3D::coords3D]; Return[$Failed]];
	
	outputlen = Length[Flatten[List[Last[temp2]]]];
	If[outputlen != 3, Message[DFInterpolation3D::nndata3D]; Return[$Failed]];
	
	centers = temp[[All,1;;inputlen]];
	fval = temp[[All,inputlen+1;;-1]];
	{centers,fval} = Developer`ToPackedArray /@ {centers, fval};
	
	If[Not[MatrixNumericQ[centers]] Not[MatrixColumnIs3[centers]] || Not[MatrixNumericQ[fval] || Not[MatrixColumnIs3[fval]] ], 
		Message[DFInterpolation::nndata3D];
		Return[$Failed]
	];
	datalen = Length[centers];
	{regularization, solvertype} = {"Regularization", "Solver"} /. Join[opts, Options[RBFInterpolation]];
	scales = GetRBFScalingParameter[centers, {}, rbf, Flatten[List[oscales]], 0.003];
	If[scales === $Failed, Return[$Failed]];
	
	(* --- Check regularization option --*)
	regularization = Flatten[List[regularization]];
	reglen = Length[regularization];
	If[reglen != 1 && reglen != datalen, 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Length[regularization]==1  && Not[NumericQ[regularization[[1]]]], 
		Message[RBFInterpolation::reg];
		regularization = {0};
	];
	If[Not[VectorNumericQ[regularization]], Message[rbf::noreg]; regularization = {0}];
	
	(* --- Check solver option --*)
	solvertype = First[Flatten[List[solvertype]]];
	If[MatchQ[solvertype, Automatic|"Automatic"], solvertype = "LinearSolve"];
	If[Not[MatchQ[solvertype, "LinearSolve" | "LeastSquares"]], 
		Message[RBFInterpolation::solver, solvertype];
		solvertype = "LinearSolve";
	];
	
	FitDivFreeRBF3D[centers, scales, fval, rbf, regularization, solvertype]
];


DivergenceFreeInterpolation3D[datapts_, opts:OptionsPattern[]] := 
								iDFRBFInterpolation3D[N@datapts, RadialBasicFunction[$DefaultRBF], Automatic, List[opts]];
DivergenceFreeInterpolation3D[datapts_, basicfun_, opts:OptionsPattern[]] := 
								iDFRBFInterpolation3D[N@datapts, iRadialBasicFunction[basicfun, "TPS2"], Automatic, List[opts]]; 
DivergenceFreeInterpolation3D[datapts_, basicfun_, rbfscales_, opts:OptionsPattern[]] :=
								iDFRBFInterpolation3D[N@datapts, iRadialBasicFunction[basicfun, "TPS2"], N@rbfscales, List[opts]];
DivergenceFreeInterpolation3D[___]:= $Failed;

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR PERFORMING COORDINATE TRANSFORMATIONS								*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

ConvertPositionFromCartesianToSigma[coords_List, zmax_?NumericQ, elevmap_Experimental`NumericalFunction]:= Module[{x,y,z, sigma},
	If[Not[MatrixNumericQ[coords]] || Not[MatrixColumnIs3[coords]] || Not[NDSolve`ValidNumericalFunctionQ[elevmap]], 
		Return[$Failed]
	];
	{x,y,z} = Transpose[coords];
	sigma = MapThread[With[{ne = elevmap[{#1, #2}]}, zmax*(#3 - ne)/(zmax - ne)]&, {x,y,z}];
	If[Not[VectorNumericQ[sigma]], $Failed, Transpose[{x,y,sigma}] ]
];

ConvertPositionFromSigmaToCartesian[sigmacoords_List, zmax_?NumericQ, elevmap_Experimental`NumericalFunction]:= Module[{x,y,sigma,z},
	If[Not[MatrixNumericQ[sigmacoords]] || Not[MatrixColumnIs3[sigmacoords]] || Not[NDSolve`ValidNumericalFunctionQ[elevmap]], 
		Return[$Failed]
	];
	{x,y,sigma} = Transpose[sigmacoords];
	z = MapThread[With[{ne = elevmap[{#1, #2}]}, (#3/zmax)*(zmax - ne) + ne]&, {x,y,sigma}];
	If[Not[VectorNumericQ[z]], $Failed, Transpose[{x,y,z}] ]
];

ConvertVelocityFromCartesianToSigma[coords_List, velocity_List, zmax_Real, elevmap_Experimental`NumericalFunction]:= 
Module[{h, dh, sigma, u,v,w, us, vs, ws},
	If[Not[MatrixNumericQ[coords]] || Not[MatrixColumnIs3[coords]] || Not[NDSolve`ValidNumericalFunctionQ[elevmap]] || 
		Not[MatrixNumericQ[velocity]] || Not[MatrixColumnIs3[velocity]], 
		Return[$Failed]
	];
	If[Dimensions[coords] =!= Dimensions[velocity], Return[$Failed]];
	{h, dh} = Transpose[Map[{elevmap[#], elevmap["Jacobian"[#]]}&, coords[[All, 1;;2]] ]];
	If[Not[VectorNumericQ[h]] || Not[MatrixNumericQ[dh]], Return[$Failed]];
	sigma = MapThread[With[{ne = elevmap[{#1, #2}]}, zmax*(#3 - ne)/(zmax - ne)]&, Transpose[coords]];
	{u,v,w} = Transpose[velocity];
	us = u*(zmax - h);
	vs = v*(zmax - h);
	ws = (zmax*w + u*(sigma - zmax)*dh[[All,1]] + v*(sigma - zmax)*dh[[All,2]]) ;
	Transpose[{us, vs, ws}]
];

ConvertVelocityFromSigmaToCartesian[sigmacoords_List, velocity_List, zmax_Real, elevmap_Experimental`NumericalFunction]:= 
Module[{h, dh, sigma, u,v,w, us, vs, ws,mask},
	If[Not[MatrixNumericQ[sigmacoords]] || Not[MatrixColumnIs3[sigmacoords]] || Not[NDSolve`ValidNumericalFunctionQ[elevmap]] || 
		Not[MatrixNumericQ[velocity]] || Not[MatrixColumnIs3[velocity]], 
		Return[$Failed]
	];
	If[Dimensions[sigmacoords] =!= Dimensions[velocity], Return[$Failed]];
	{h, dh} = Transpose[Map[{elevmap[#], elevmap["Jacobian"[#]]}&, sigmacoords[[All, 1;;2]] ]];
	If[Not[VectorNumericQ[h]] || Not[MatrixNumericQ[dh]], Return[$Failed]];
	sigma = sigmacoords[[All,-1]];
	mask = Boole[Thread[sigma > -0.0003]];
	{us,vs,ws} = Transpose[velocity];
	u = mask*us/(zmax - h);
	v = mask*vs/(zmax - h);
	w = mask*(ws - u*(sigma - zmax)*dh[[All,1]] - v*(sigma - zmax)*dh[[All,2]])/zmax ; 
	Transpose[{u, v, w}]
];

(* =====================================================================================================*)
(* 						FUNCTIONS FOR IMPORTING AND PROCESSING THE SDGE DATA							*)
(* =====================================================================================================*)

$SENSORELEVATION = 32.8; (* The velocity sensors are 32.8ft above ground *)

LoadSDGEWeatherStationInfo[]:= Module[{},
	SetDirectory[$SDGEDATAPATH];
	Get["weatherStationInfo.mx"];
];

LoadSDGEWeatherData[filename_, n_Integer] := Module[{temp1, nmax = 52845},
  If[n > nmax, Return[$Failed]];
  temp1 = ReadList[filename, Record, n, RecordSeparators -> {"\r\n"}, RecordLists -> True];
  ToExpression[ToString[temp1[[n]]] ]
  ];

Options[LoadSDGESensorLocations] = {"LLECoordinates" -> False};  
LoadSDGESensorLocations[stations_, OptionsPattern[]]:= Module[{datafilenames, stationCodes, nameCode, dataNotAvailable, CoordsNotAvailable,
	dataAndCoordsAvailable, dataindex, stationindex, LongLatElevCoords, opt},
	SetDirectory[$SDGEDATAPATH];
	datafilenames = FileNames["*_2012.dat"] ;
	stationCodes = stations[[2 ;; -1, 2]] ;
	nameCode = Map[ToUpperCase[StringTake[#, 3]] &, datafilenames] ; 
	dataNotAvailable = Complement[stationCodes, nameCode] ; 
	CoordsNotAvailable = Complement[nameCode, stationCodes];
	dataAndCoordsAvailable = Intersection[nameCode, stationCodes];
	
	dataindex = Map[First@Flatten@Position[nameCode, #] &, dataAndCoordsAvailable] ; 
	stationindex = Map[First@Flatten@Position[stationCodes, #] &, dataAndCoordsAvailable];
	LongLatElevCoords = ToExpression[stations[[stationindex, {5, 4, 7}]]] ;
	
	opt = OptionValue["LLECoordinates"];
	{If[opt, LongLatElevCoords, LongLatElevToCartesianCoordinates[LongLatElevCoords]], dataindex}
];

LoadSDGEVelocityData[dayindex_, dataindex_]:= Module[{datafilenames, stationData},
	SetDirectory[$SDGEDATAPATH];
	datafilenames = FileNames["*_2012.dat"];
	stationData = Map[LoadSDGEWeatherData[datafilenames[[#]], dayindex] &, dataindex];
	stationData[[All, {5, 6}]] /. {speed_, direction_} :> -speed*{Sin[direction*Pi/180.], Cos[direction*Pi/180.],0.}
];

LongLatElevToCartesianCoordinates[llecoords_List]:= Module[{long, lat, elev,x,y},
	If[Not[MatrixNumericQ[llecoords]], Return[$Failed]];
	If[Not[TrueQ[Last[Dimensions[llecoords]] == 3]], Return[$Failed]];
	{long, lat, elev} = Transpose[llecoords];
	x = $RadiusOfEarthInFeet*Cos[lat*$da]*Cos[long*$da];
	y = $RadiusOfEarthInFeet*Cos[lat*$da]*Sin[long*$da];
	Transpose[{x,y,elev}]
];

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						FUNCTIONS FOR PERFORMING THE LEVEL-SET SIMULATION								*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

GridPoints[{xmin_, xmax_, nx_Integer}, {ymin_, ymax_, ny_Integer}] := Module[ {gridpts, xgrid, ygrid},
        xgrid = N@Range[xmin, xmax, (xmax - xmin)/nx];
        ygrid = N@Range[ymin, ymax, (ymax - ymin)/ny];
        gridpts = Flatten[Outer[List, xgrid, ygrid], 1];
        {xgrid, ygrid, gridpts}
    ];

InterpolatingFunctionFromData[data_?VectorQ, TransportData_] := Module[ {},
        Interpolation[Join[TransportData["Grid"], Transpose[{data}], 2], InterpolationOrder->2]
    ];

PopulateData[{xmin_, xmax_, nx_Integer}, {ymin_, ymax_, ny_Integer}, TransportData_] := 
	Module[ {xg, yg, eye, ddx, ddy, d2x, d2y, d2dxdy, grid, bcposDDX, bcposDDY, massMatrix, x, y, allbcpos},
        {xg, yg, grid} = GridPoints[{xmin, xmax, nx}, {ymin, ymax, ny}];
        {eye, ddx, ddy, d2x, d2y,d2dxdy} = NDSolve`FiniteDifferenceDerivative[#, {xg, yg}, 
        	"DifferenceOrder" -> 4]["DifferentiationMatrix"] & /@ {{0, 0}, {1, 0}, {0, 1}, {2, 0}, {0, 2}, {1,1}};
        bcposDDX = DeleteDuplicates@Flatten[Position[grid, #] & /@ {{xmin, y_}, {xmax, y_}}];
        bcposDDY = DeleteDuplicates@Flatten[Position[grid, #] & /@ {{x_, ymin}, {x_, ymax}}];
        
        TransportData["Xgrid"] = xg;
        TransportData["Ygrid"] = yg;
        TransportData["Grid"] = grid;
        TransportData["DifferentiationMatrices"] = {ddx, ddy, d2x, d2y, d2dxdy};
        TransportData["Constants"] = {0.165, 3.24, 3/4, 0.5};
        
        (* Mass matrix for explicit methods *)
        massMatrix = eye;
        massMatrix[[bcposDDX]] = ddx[[bcposDDX]];
        massMatrix[[bcposDDY]] = ddy[[bcposDDY]];
        allbcpos = Union[bcposDDX, bcposDDY];
        
        TransportData["BCPositions"] = allbcpos;
        TransportData["MassMatrixFunction"] = LinearSolve[massMatrix];
    ]; 

TransportRHS[c_?VectorQ, t_?NumericQ, TransportData_] := 
	Module[ {r0, cf, n, alp, dcdx, dcdy, d2cdx2, d2cdy2, d2cdxdy, gnorm, nx, ny, vnorm, vn, theta, c1, c2, f, 
		ux, uy, rhs, vx, vy, dd, lengthscale},
      
      	{r0, cf, n, alp} = TransportData["Constants"];
      	lengthscale = TransportData["LengthScale"];
      	
        {dcdx, dcdy, d2cdx2, d2cdy2, d2cdxdy} = (#.c) & /@ TransportData["DifferentiationMatrices"];
        gnorm = Sqrt[dcdx^2 + dcdy^2 + 10^-15];
        {nx, ny} = {dcdx/gnorm, dcdy/gnorm};
       	{vx, vy} = TransportData /@ {"XVelocity", "YVelocity"};
        vnorm = Sqrt[vx^2 + vy^2];
        vn = vx*nx + vy*ny;
        theta = ArcCos[vn/vnorm];
        c1 = r0*(1 + cf*Sqrt[vnorm]*(Cos[theta]^2)^n);
        c2 = r0*(alp + (1 - alp)*Sin[theta]);
        f = Chop[Boole[Thread[Abs[theta] <= Pi/2]]*c1 + Boole[Thread[Pi/2 < Abs[theta] <= Pi]]*c2];
        {ux, uy} = {f*nx, f*ny}/lengthscale;
        
        (*dd = Sqrt[dcdx^2 + dcdy^2]; dd = TransportData["Diffusion"]*dd/Max[dd];*)
        (*dd = (d2cdx2*dcdy^2 - 2*d2cdxdy*dcdx*dcdy + d2cdy2*dcdx^2)/(dcdx^2 + dcdy^2 + 10^-15)^(3/2);
        dd = dd^2; dd = TransportData["Diffusion"]*(1 + dd/Max[dd]);*)
        dd = TransportData["Diffusion"]/lengthscale;
        
        rhs = -ux*dcdx - uy*dcdy + dd*(d2cdx2 + d2cdy2);
        rhs[[TransportData["BCPositions"] ]] = 0.;
        rhs = TransportData["MassMatrixFunction"][rhs];
        rhs
    ];

RK4TimeStep[ic_?VectorQ, t_?NumericQ, TransportData_] := Module[ {k0, k1, k2, k3, dh},
		dh = TransportData["TimeStep"];
        k0 = dh*TransportRHS[ic, t, TransportData];
        k1 = dh*TransportRHS[ic + k0/2, t + dh/2, TransportData];
        k2 = dh*TransportRHS[ic + k1/2, t + dh/2, TransportData];
        k3 = dh*TransportRHS[ic + k2, t + dh, TransportData];
        ic + (k0 + 2*k1 + 2*k2 + k3)/6
    ];

TransportIterate[ic_?VectorQ, dtMax_?NumericQ, timeInfo_List, TransportData_] :=
	Module[ {dt = dtMax, time, i = 2, sol = ic, saveflag = False, timesol, maxiter, iter},
        time = First[timeInfo];
        maxiter = Ceiling[Last[timeInfo]/dtMax] + 1;
        timesol = InterpolatingFunctionFromData[ic, TransportData];
        timesol = Flatten[List[timesol]];
        TransportData["TimeStep"] = dtMax;
        Do[
         If[ (time + dt) >= timeInfo[[i]],
             dt = timeInfo[[i]] - time;
             saveflag = True;
             i++;
             ,
             dt = dtMax;
         ];
         sol = RK4TimeStep[sol, time, TransportData];
         sol = (sol/Max[Abs[sol]]);
         If[ saveflag,
             timesol = Join[timesol, {InterpolatingFunctionFromData[sol, TransportData]}];
             saveflag = False;
             If[ i > Length[timeInfo], Break[]];
         ];
         time += dt;
         , {iter, maxiter}
         ];
        timesol
    ];

$DefaultDiffusion = 0.001;    
$DefaultGridPoints = {100,100};
TrackLevelSet[initialCondition_, domain_List, timestep_?NumericQ, savetimes_List, velocityinfo_List, opts:OptionsPattern[]] := Module[
	{diffusion, gridpoints, lengthscale, xinfo, yinfo, unf, vnf, TransportData, saveinfo, temp, ic},
	
	{diffusion, gridpoints, lengthscale} = {"Diffusion", "nGridPoints", "LengthScale"} /. Join[List[opts], Options[TrackLevelSet]];
	diffusion = First[Flatten[List[diffusion]]]; 
	If[MatchQ[diffusion, Automatic | "Automatic"], diffusion = $DefaultDiffusion ];
	If[Not[NumericQ[diffusion]], diffusion = $DefaultDiffusion];
	
	gridpoints = Flatten[List[gridpoints]];
	If[MatchQ[First[gridpoints] , Automatic | "Automatic"], gridpoints = $DefaultGridPoints];
	If[Not[VectorIntegerQ[gridpoints] && Length[gridpoints]==2], gridpoints = $DefaultGridPoints];
	
	lengthscale = First[Flatten[List[lengthscale]]];
	If[MatchQ[lengthscale, Automatic | "Automatic"], lengthscale = 1];
	If[Not[NumericQ[lengthscale]], lengthscale = 1];
	
	saveinfo = Flatten[List[savetimes]];
	If[Not[VectorNumericQ[saveinfo]], Return[$Failed]];
	saveinfo = Sort[saveinfo];
	
	TransportData["Diffusion"] = diffusion;
	TransportData["LengthScale"] = lengthscale;
	
	If[Not[MatrixNumericQ[domain] && Dimensions[domain] == {2,2}], Return[$Failed]];
	{xinfo,yinfo} = Map[Flatten, Join[domain, Transpose[{gridpoints}], 2] ];
	
	(* --- the velocity must be in the form of a numerical function ---*)
	If[Length[velocityinfo] != 2, Return[$Failed]];
	If[Not[And@@Map[NDSolve`ValidNumericalFunctionQ, velocityinfo]], Return[$Failed]];
	{unf, vnf} = velocityinfo;
	
	PopulateData[xinfo, yinfo, TransportData];
	TransportData["XVelocity"] = TransportData["Grid"] /. {x$_,y$_} :>  unf[{x$,y$,0.}] ; 
	TransportData["YVelocity"] = TransportData["Grid"] /. {x$_,y$_} :> vnf[{x$,y$,0.}] ; 
	
	If[Head[initialCondition] === Experimental`NumericalFunction, 
		If[Length[initialCondition["ArgumentNames"]] != 2, Return[$Failed]];
        ic = Map[initialCondition, TransportData["Grid"]]; 
        ,
        temp = TransportData["Grid"][[1]];
		If[Not[NumericQ[initialCondition[Sequence@@temp]]], Return[$Failed]];
        ic = Map[initialCondition[Sequence@@#]&, TransportData["Grid"]];
	];
	If[Not[VectorNumericQ[ic]], Return[$Failed]];
	
	Print["iterating...."];
	
	TransportIterate[ic, timestep, saveinfo, TransportData]
]; 

Wildfires`Utility`PlotZeroLevelSetsInSigmaPlane[tinfo_List, ss_List, {x_Symbol,xmin_?NumericQ, xmax_?NumericQ},
	{y_Symbol, ymin_?NumericQ, ymax_?NumericQ}] := Module[{cp1, ti, cplot, si},
	
	ti = 1;
	cp1 = ContourPlot[ss[[1]][x, y] == 0, {x, xmin, xmax}, {y, ymin, ymax}, Frame -> True, GridLines -> Automatic, 
		AspectRatio -> Automatic, ContourStyle -> {Thick, ColorData[{"FallColors", "Reverse"}][ti]}, 
   		PlotLegends -> BarLegend[{{"FallColors", "Reverse"}, {0, tinfo[[-1]]}}, 
     	LegendMarkerSize -> 550, LabelStyle -> Directive[Black, FontSize -> 16]], 
   		ImageSize -> 600
   	];
   	cplot = Show[cp1, Table[
   		ti = tinfo[[si]]/(tinfo[[-1]]);
   		ContourPlot[ss[[si]][x, y] == 0, {x, xmin, xmax}, {y, ymin,ymax}, Frame -> True, GridLines -> Automatic, 
    			AspectRatio -> Automatic, ContourStyle -> {Thick, ColorData[{"FallColors", "Reverse"}][ti]}], 
    				{si, 2, Length[tinfo]}
   	],
   	FrameLabel -> {Style[Text["x"], Black, 20], Style[Text["y"], Black, 20]}, FrameStyle -> Directive[Black, 16]
  ]; 
  cplot
];

Wildfires`Utility`PlotZeroLevelSetsInCartesianDomain[tinfo_List, ss_List, nfelev_Experimental`NumericalFunction, $ZMAX_, 
	{x_Symbol,xmin_?NumericQ, xmax_?NumericQ}, {y_Symbol, ymin_?NumericQ, ymax_?NumericQ}] := Module[
	{popts, nfe, ti, plotfunc, mo, coords, contours,cc,ccXYZLine, si},
	
	popts = {PlotPoints -> Automatic, "BoundaryOffset" -> False, "ReturnRawMeshObject" -> True, MaxRecursion -> Automatic};
	nfe = Plot3D[nfelev[{x, y}], {x, xmin, xmax}, {y, ymin, ymax}, Mesh -> None, ColorFunction -> "CherryTones", PlotRange -> All, 
   				Filling -> Bottom, FillingStyle -> Black, PlotPoints -> 100, ImageSize -> 700];
   Show[nfe, Table[
   	ti = tinfo[[si]]/(tinfo[[-1]]);
  	plotfunc = Join[{ss[[si]][x, y] == 0.0, {x, xmin, xmax}, {y, ymin, ymax}, popts}];
  	mo = Apply[Visualization`Core`ContourPlot, plotfunc];
  	{coords, contours} = Map[(mo@#) &, {"VertexCoordinates", "MeshLinesVertices"}];
  	cc = Transpose[Join[Transpose[coords], 0*{coords[[All, 1]]} ]]; 
  	cc = ConvertPositionFromSigmaToCartesian[cc, $ZMAX, nfelev];
  	ccXYZLine = Line[cc[[#]]] & /@ contours ;
  	Graphics3D[{Thickness[0.005], ColorData[{"AvocadoColors", "Reverse"}][ti], ccXYZLine}]
  	, {si, 1, Length[tinfo]}]
  	,
  	BoxRatios -> {1, 1, 0.25},AxesLabel -> {Style[Text["x"], Black, Thick, 17], 
   	Style[Text["y"], Black, Thick, 17], Style[Text["z  "], Black, Thick, 17]}, 
 	AxesStyle -> Directive[Black, Thick, 16], Axes -> True
 	] 
];

(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* 						UTILITY FUNCTIONS																*)
(* =====================================================================================================*)
(* =====================================================================================================*)
(* =====================================================================================================*)

Extract2DIsoSurfacePoints[fun_, zlevel_, {x_, xmin_, xmax_}, {y_, ymin_, ymax_}, npts_Integer] :=
    Module[ {popts, refopts, plotfunc, coords, contours, pp, edgoal, slgoal, mo, newpts, abscissa, if, n, pi},
        popts = {PlotPoints -> Automatic, "BoundaryOffset" -> False, "ReturnRawMeshObject" -> True, MaxRecursion -> Automatic};
        refopts = {"Divisions" -> 100, "VertexTolerance" -> 10^(-13)};
        popts = Append[popts, "Refinement" -> refopts];
        (*If[Not[Head[fun]===Function || Head[fun]===Experimental`NumericalFunction], Return[$Failed]];*)
        If[Head[fun] =!= Experimental`NumericalFunction, 
        	If[Not[NumericQ[fun[xmin,ymin]]], Return[$Failed]];
        	
        	plotfunc = Join[{fun[x,y] == zlevel, {x, xmin, xmax}, {y, ymin, ymax}}, popts]
        	,
        	If[Length[fun["ArgumentNames"]] != 2, Return[$Failed]];
        	plotfunc = Join[{fun[{x,y}] == zlevel, {x, xmin, xmax}, {y, ymin, ymax}}, popts]
        ];
       	mo = Apply[Visualization`Core`ContourPlot, plotfunc];
        If[ mo === $Failed, Return[$Failed]];
        {coords, contours} = Map[(mo@#) &, {"VertexCoordinates", "MeshLinesVertices"}];
        newpts = Table[
          pp = coords[[contours[[n]] ]];
          edgoal = Norm[Subtract @@ Transpose[{Min[#], Max[#]} & /@ Transpose[pp]]]/2500;
          slgoal = Last[Accumulate[Map[Norm, Differences[pp]]]];
          First[ Graphics`Mesh`SimplifyPolyLine[Line[pp], IntersectionsFree -> True, SegmentPartition -> False, 
            EdgesDistanceGoal -> edgoal, SegmentLengthGoal -> slgoal]] 
          ,
          {n, Length[contours]}
          ];
        newpts = Table[
          abscissa = Prepend[Accumulate[Map[Norm, Differences[pi]]] , 0.];
          if = Interpolation[Transpose[{abscissa, pi}], InterpolationOrder -> 2];
          With[ {tr = First[if["Domain"]]}, DeleteDuplicates[if[Range @@ Flatten[List[tr, Last[tr]/npts]]] ]
          ]
          , {pi, newpts}
          ];
        newpts
    ]

ReOrderPoints3D[p1_, p2_, v1_, v2_] := Block[{temp = ConstantArray[0, 3]}, temp[[p1]] = v1; temp[[p2]] = v2; temp] ; 
ReOrderPoints2D[p1_, p2_, v1_, v2_] := Block[{temp = ConstantArray[0, 2]}, temp[[p1]] = v1; temp[[p2]] = v2; temp] ; 
  
Extract3DVelocityDataFromBoundaryConditions[bcs_List, {xg_List, yg_List, zg_List}] := 
 Module[{xyzgrid, data, bci, bcfun, allgrid, pointgrid, pp, pointval, p2, i},
  xyzgrid = {xg, yg, zg};
  If[Not[And @@ Map[Length[#] == 2 &, bcs]], Return[$Failed]];
  If[Not[And @@ Map[Length[First[#]] == 3 &, bcs]], Return[$Failed]];
  If[Not[And @@ Map[Head[Last[#]] === Function &, bcs]], Return[$Failed]];
  data = Table[
  	{bci, bcfun} = bcs[[i]];
    allgrid = Pick[Range[3], Map[Head[#] === Symbol &, bci]];
    pointgrid = Complement[Range[3], allgrid];
    If[allgrid =!= {},
     pp = Flatten[Outer[List, Sequence @@ xyzgrid[[allgrid]]], 1] ; 
     pointval = bci[[pointgrid]];
     p2 = Map[ReOrderPoints3D[allgrid, pointgrid, #, pointval] &, pp] ; 
     ,
     p2 = pointgrid;
     ];
    If[Not[MatrixNumericQ[p2] && MatrixColumnIs3[p2]], Return[$Failed]];
    p2 /. {x$_?NumericQ, y$_?NumericQ, z$_?NumericQ} :> {{x$,y$,z$}, bcfun[x$, y$, z$]}
    , {i, Length[bcs]}
    ];
  Join@@data
  ]

Extract2DVelocityDataFromBoundaryConditions[bcs_List, {xg_List, yg_List}] := 
 Module[{xygrid, data, bci, bcfun, allgrid, pointgrid, pp, pointval, p2, i},
  xygrid = {xg, yg};
  If[Not[And @@ Map[Length[#] == 2 &, bcs]], Return[$Failed]];
  If[Not[And @@ Map[Length[First[#]] == 2 &, bcs]], Return[$Failed]];
  If[Not[And @@ Map[Head[Last[#]] === Function &, bcs]], Return[$Failed]];
  data = Table[
  	{bci, bcfun} = bcs[[i]];
    allgrid = Pick[Range[2], Map[Head[#] === Symbol &, bci]];
    pointgrid = Complement[Range[2], allgrid];
    If[allgrid =!= {},
     pp = Flatten[Outer[List, Sequence @@ xygrid[[allgrid]]], 1] ; 
     pointval = bci[[pointgrid]];
     p2 = Map[ReOrderPoints2D[allgrid, pointgrid, #, pointval] &, pp] ; 
     ,
     p2 = pointgrid;
     ];
    If[Not[MatrixNumericQ[p2] && MatrixColumnIs2[p2]], Return[$Failed]];
    p2 /. {x$_?NumericQ, y$_?NumericQ} :> {{x$,y$}, bcfun[x$, y$]}
    , {i, Length[bcs]}
    ];
  Join@@data
  ]


Wildfires`Utitlity`VelocityDataFromTerrainAndBCs[terrain_, bcs_List , {x_Symbol, xmin_?NumericQ, xmax_?NumericQ}, {y_Symbol, ymin_?NumericQ, ymax_?NumericQ}, 
	{z_Symbol, zmin_?NumericQ, zmax_?NumericQ}, {nx_Integer, ny_Integer, nz_Integer}] := Module[
		{xg, yg, zg, xygrid, data, terraindata},
		
		If[Not[Head[terrain]===Function || Head[terrain]===Experimental`NumericalFunction], Return[$Failed]];
		
		xg = N@Range[xmin, xmax, Abs[(xmax - xmin)]/nx];
		yg = N@Range[ymin, ymax, Abs[(ymax - ymin)]/ny]; 
		zg = N@Range[zmin, zmax, Abs[(zmax - zmin)]/nz];
		If[xg === {} || yg === {} || zg === {}, Return[$Failed]];
		
		data = Extract3DVelocityDataFromBoundaryConditions[bcs, {xg[[2;;-1]], yg[[2;;-1]], zg[[2;;-2]]}];
		If[data === $Failed, Return[$Failed]];
		
		xg = N@Range[xmin, xmax, Abs[(xmax - xmin)]/30];
		yg = N@Range[ymin, ymax, Abs[(ymax - ymin)]/30]; 
		xygrid = Flatten[Outer[List, xg, yg],1];
		If[Head[terrain] === Function,
			If[Not[NumericQ[terrain[xmin,ymin]]], Return[$Failed]];
			terraindata = xygrid /. {x$_?NumericQ, y$_?NumericQ}:> {{x$,y$,terrain[x$,y$]}, {0.,0.,0.}} ;
			,
			terraindata = Join[xygrid, Transpose[List[terrain[xygrid]]] , 2];
		];
		
		Join[data, terraindata]
	];

Wildfires`Utility`VelocityDataInXYPlane[terrain_,bcs_List, {x_Symbol, xmin_?NumericQ, xmax_?NumericQ}, {y_Symbol, ymin_?NumericQ, ymax_?NumericQ}, 
	zlevel_?NumericQ, {nx_Integer, ny_Integer}, rbf_, scale_]:= Block[{xg, yg, grid, sol, data, terrainPoints},
	
	If[Not[Head[terrain]===Function || Head[terrain]===Experimental`NumericalFunction], Return[$Failed]];
	xg = N@Range[xmin, xmax, Abs[(xmax - xmin)]/nx];
	yg = N@Range[ymin, ymax, Abs[(ymax - ymin)]/ny];
	If[xg === {} || yg === {}, Return[$Failed]];
	
	data = Extract2DVelocityDataFromBoundaryConditions[bcs, {xg[[2;;-1]], yg[[2;;-1]]}];
	If[data === $Failed, Return[$Failed]];
	
	terrainPoints = Flatten[Extract2DIsoSurfacePoints[terrain, zlevel, {x, xmin, xmax}, {y, ymin, ymax}, 30] , 1] ;
	If[terrainPoints===$Failed, Return[$Failed]];
	terrainPoints = Transpose[{terrainPoints, N[0*terrainPoints]}];
	
	data = Join[data, terrainPoints];
	sol = DivergenceFreeInterpolation2D[data, rbf, scale];
	If[sol === $Failed, Return[$Failed]];
	grid = Flatten[Outer[List, xg, yg], 1];
	Transpose[{grid, sol[grid]}]
]; 

SingleHill[x_,y_]:= Exp[-0.02*((x - 3)^2 + y^2)];
TwoHills[x_,y_] := Exp[-0.02*((x - 3)^2 + y^2)] + Exp[-0.05*((x - 9)^2 + (y + 9)^2)] ;
ThreeHills[x_,y_]:= Exp[-0.02*((x - 3)^2 + y^2)] + Exp[-0.05*((x - 9)^2 + (y + 9)^2)] + Exp[-0.05*((x - 8)^2 + (y - 8)^2)];

MakeHexGrid[{xmin_, xmax_}, {ymin_, ymax_}, dh_] := Module[{grid1, grid2}, 
  grid1 = Flatten[Outer[List, Range[xmin, xmax, dh], Range[ymin, ymax, Sqrt[3]*dh]], 1];
  grid2 = grid1 /. {x_, y_} :> {x + dh/2, y + (Sqrt[3]*dh/2)};
  Flatten[{grid1, grid2}, 1]
]

MakeSquareGrid[{xmin_,xmax_}, {ymin_, ymax_}, dh_]:= Flatten[Outer[List, Range[xmin,xmax,dh], Range[ymin, ymax,dh]], 1];

$GROUNDSENSORX = 30;
$GROUNDSENSORY = 30;
Wildfires`Utility`DiscretizeElevationMap[elevmap_,{xmin_?NumericQ,xmax_?NumericQ, nx_Integer}, {ymin_?NumericQ, ymax_?NumericQ, ny_Integer}]:= Module[
	{groundsensorsXYZ, xygrid, dh},
	dh = Min[(xmax-xmin)/nx, (ymax-ymin)/ny];
	xygrid = MakeHexGrid[{xmin, xmax}, {ymin, ymax}, dh];
	If[Head[elevmap] === Experimental`NumericalFunction, 
		If[Length[elevmap["ArgumentNames"]] != 2, Return[$Failed]];
        groundsensorsXYZ = Map[Flatten[{#,elevmap[#]}]&, xygrid]; 
        ,
        If[Not[NumericQ[elevmap[xmin,ymin]]], Return[$Failed]];
        groundsensorsXYZ = Map[Flatten[{#,elevmap[Sequence@@#]}]&, xygrid];
	];
	groundsensorsXYZ
];

iGetHillVelocityField[hills_]:= Module[{zi, uvel, xmin,xmax,ymin,ymax,zmin,zmax, sensor2DLocation1, sensor2DLocation2, 
	velInfo2D1, velInfo2D2, sensors2D, velocity2D, sensorlocations0, sensorlocations1, sensorlocations2, sensorlocations3, 
	uvwdata0, uvwdata1, uvwdata2, uvwdata3, x,y, rbfvel2D, data, domain, rbfsol3D},
	uvel = Function[{z$}, 3*Erf[z$]];
	zi = 0.7;
	{{xmin,xmax},{ymin,ymax},{zmin,zmax}} = {{-15,20},{-20,20},{0,2}};
	domain  ={{xmin,xmax},{ymin,ymax},{zmin,zmax}};
	
	sensor2DLocation1 = Flatten[Extract2DIsoSurfacePoints[hills, zi, {x, xmin, xmax}, {y, ymin, ymax}, 30] , 1] ; 
	sensor2DLocation2 = Thread[{xmin, N@Range[ymin, ymax, (ymax - ymin)/40]}];
	velInfo2D1 = 0.*sensor2DLocation1;
	velInfo2D2 = sensor2DLocation2 /. {x$_, y$_} :> {uvel[zi], 0.} ;
	sensors2D = Join[sensor2DLocation1, sensor2DLocation2];
	velocity2D = Join[velInfo2D1, velInfo2D2];
	rbfvel2D = DivergenceFreeInterpolation2D[Transpose[{sensors2D, velocity2D}], "MQ", 0.6];
	
	sensorlocations0 = Flatten[Outer[List, Range[xmin + 1, xmax, (xmax - (xmin + 1))/10], Range[ymin, ymax, (ymax - ymin)/10]], 1] ;
	uvwdata0 = Transpose[List[sensorlocations0, rbfvel2D[sensorlocations0]]];
	uvwdata0 = uvwdata0 /. {{x$_,y$_},{u$_,v$_}} :> {{x$,y$,zi},{u$,v$,0.}};
	
	sensorlocations1 = Flatten[Outer[{xmin, #1, #2} &, Range[ymin, ymax, (ymax - ymin)/10], Range[(zmin + 0.01), zmax, (zmax - (zmin + 0.01))/10]], 1] ; 
	uvwdata1 = sensorlocations1 /. {x$_, y$_, z$_} :> N[{{x$,y$,z$},{uvel[z$], 0, 0}}];
	
	sensorlocations2 = Flatten[Outer[{#1, #2, zmax} &, Range[xmin + 1, xmax, (xmax - (xmin + 1))/10], Range[ymin, ymax, (ymax - ymin)/10]], 1] ;
	uvwdata2 = sensorlocations2 /. {x$_, y$_, z$_} :> N[{{x$,y$,z$},{uvel[z$], 0, 0}}];
	
	sensorlocations3 = Wildfires`Utility`DiscretizeElevationMap[hills, {xmin,xmax,30},{ymin,ymax,30}];
	uvwdata3 = Transpose[{sensorlocations3, 0.*sensorlocations3}];
	
	data = Join[uvwdata0, uvwdata1, uvwdata2, uvwdata3];
	rbfsol3D = DivergenceFreeInterpolation3D[data, "TPS2"];
	{Evaluate[hills[#1,#2]]&, 1.1*zmax, domain, data, rbfsol3D}
];

Wildfires`Utility`TestCase[name_String]:= Module[{sname}, 
	sname = Switch[name, "SingleHill", SingleHill, "TwoHills", TwoHills, "ThreeHills", ThreeHills,_,Return["NotAvailable"]];
	iGetHillVelocityField[sname]
]; 
Wildfires`Utility`TestCase["Names"] := {"SingleHill", "TwoHills", "ThreeHills"};
Wildfires`Utility`VelocityTestCase[___]:= $Failed;

End[]

EndPackage[]

